from skretrieval.retrieval import RetrievalTarget
import numpy as np
import sasktran as sk
from skretrieval.core.radianceformat import RadianceBase, RadianceGridded
from copy import copy
import logging


class OSIRISAlbedoRetrieval(RetrievalTarget):
    def __init__(self, atmosphere: sk.Atmosphere, high_alt: float,  ret_wavel, ref_range=5000, use_aerosol_reference_altitude=False,
                 aerosol_ref_range=(30000, 40000)):
        self._atmosphere = atmosphere
        self._ret_wavel = ret_wavel
        self._high_alt = high_alt
        self._ref_range = ref_range
        self._albedo = copy(atmosphere.brdf.albedo)
        self._smallest_albedo = 1e-3
        self._largest_albedo = 1
        self._zero_meas_vector = 0
        self._use_aerosol_reference_altitude = use_aerosol_reference_altitude
        self._aerosol_ref_range = aerosol_ref_range

    def configure_from_model(self, forward_model, measurement_l1):
        old_brdf = copy(self._atmosphere.brdf)
        # Set the albedo to 0 to get the meas vector with albedo 0
        self._atmosphere.brdf = 0.0
        rad = forward_model.calculate_radiance()
        self._atmosphere.brdf = old_brdf

        if self._use_aerosol_reference_altitude:

            if 'aerosol' in forward_model.engine.atmosphere.species.keys():
                spec = forward_model.engine.atmosphere.species['aerosol'].species
                old_aerosol = copy(forward_model.engine.atmosphere.species['aerosol'].climatology[spec])
                forward_model.engine.atmosphere.species['aerosol'].climatology[spec] = np.zeros_like(old_aerosol)

            rayleigh_radiance = forward_model.calculate_radiance()

            if 'aerosol' in forward_model.engine.atmosphere.species.keys():
                spec = forward_model.engine.atmosphere.species['aerosol'].species
                forward_model.engine.atmosphere.species['aerosol'].climatology[spec] = old_aerosol

            meas_w = rayleigh_radiance.data.wavelength.values[0]
            tangents = measurement_l1.tangent_locations()
            aer_y = np.log(measurement_l1.data.sel(wavelength=meas_w)) - np.log(rayleigh_radiance.data)
            aer_y['tangent_altitude'] = (['los'], tangents['altitude'].values)
            aer_y = aer_y.swap_dims({'los': 'tangent_altitude'})
            aer_y = aer_y.where((aer_y.tangent_altitude > self._aerosol_ref_range[0]) & (aer_y.tangent_altitude < self._aerosol_ref_range[1]), drop=True)

            min_idx = np.nanargmin(aer_y['radiance'])
            self._high_alt = aer_y.tangent_altitude.values[min_idx] + 50
            self._ref_range = 100

        self._zero_meas_vector = self.measurement_vector(rad)

    def measurement_vector(self, l1_data: RadianceBase):
        if not isinstance(l1_data, RadianceGridded):
            raise ValueError('Class Albedo Retrieval only supports data in the form RadianceGridded')

        result = dict()

        tangent_locations = l1_data.tangent_locations()
        l1_data.data['tangent_altitude'] = tangent_locations.altitude

        triplet_values = l1_data.data['radiance'].where((l1_data.data['tangent_altitude'] > self._high_alt - self._ref_range) & (l1_data.data['tangent_altitude'] < self._high_alt), drop=False).sel(wavelength=self._ret_wavel, method='nearest')

        result['y'] = float(np.log(triplet_values).mean(dim='los').values)

        if self._zero_meas_vector:
            result['jacobian'] = (result['y'] - self._zero_meas_vector['y']) / float(self.state_vector())
            result['jacobian'] = result['jacobian'].reshape(1, 1)

        result['y'] = np.array([result['y']])

        return result

    def state_vector(self):
        return np.array([self._albedo])

    def update_state(self, x: np.ndarray):
        self._albedo = np.nanmax([self._smallest_albedo, copy(float(x))])
        self._albedo = np.nanmin([self._largest_albedo, self._albedo])
        self._atmosphere.brdf = copy(self._albedo)
        logging.info("Setting albedo to: {}".format(self._albedo))

    def apriori_state(self):
        return None

    def inverse_apriori_covariance(self):
        return None

    def measurement_wavelengths(self, measurement_l1):
        wavel = measurement_l1.data.sel(wavelength=self._ret_wavel, method='nearest')['wavelength'].values

        return np.array(wavel)

    def rtm_options(self):
        options = dict()
        options['numordersofscatter'] = 50

        return options