from osiris_skretrieval.osiris import Osiris
import sasktran as sk
import numpy as np
from scipy import interpolate
import skretrieval.util as util
from skretrieval.core.lineshape import DeltaFunction, Rectangle
from skretrieval.core.sensor.spectrograph import SpectrographFast
import skretrieval.core.radianceformat as radianceformat
import xarray as xr


class OSIRISModel(object):
    def __init__(self, osiris: Osiris, atmosphere: sk.Atmosphere, model_vertical_resolution_rad=0.0002, options=None, model_wavel=None,
                 meas_wavel=None):
        # Calculate the scan rate in rad/s
        geodetic = sk.Geodetic()
        self.measurement_optical_axis = osiris.optical_axis()
        self.measurement_l1 = osiris.skretrieval_l1()
        tangent_locations = self.measurement_l1.tangent_locations()
        tangent_angles = []
        tangent_mjd = []
        for idx, o in enumerate(self.measurement_optical_axis):
            tangent = tangent_locations.isel(los=idx)

            geodetic.from_lat_lon_alt(float(tangent.latitude.values), float(tangent.longitude.values), float(tangent.altitude.values))

            tangent_radius = np.linalg.norm(geodetic.location)

            obs_radius = np.linalg.norm(o.observer)

            tangent_angles.append(np.arcsin(tangent_radius / obs_radius))
            tangent_mjd.append(o.mjd)

        scan_rate = np.polyfit(tangent_mjd, tangent_angles, deg=1)[0]

        self.scan_rate = np.abs(scan_rate / (24 * 3600)) # To rad/s

        # Set up high resolution vertical geometry
        self.measurement_geometry = osiris.sasktran_geometry()
        self.model_geometry = sk.Geometry()

        lower_los = self.measurement_geometry.lines_of_sight[0]
        upper_los = self.measurement_geometry.lines_of_sight[-1]

        rotation_axis = np.cross(lower_los.look_vector, upper_los.look_vector)
        rotation_axis /= np.linalg.norm(rotation_axis)

        # Need to get the measurement angles to interpolate the mjd's properly
        all_angles = np.array([np.arccos(np.dot(l.look_vector, lower_los.look_vector)) for l in self.measurement_geometry.lines_of_sight])
        all_angles[0] = 0
        all_mjd = np.array([l.mjd for l in self.measurement_geometry.lines_of_sight])
        all_observer = np.array([l.observer for l in self.measurement_geometry.lines_of_sight])

        mjd_interpolator = interpolate.interp1d(all_angles, all_mjd, bounds_error=False, fill_value='extrapolate')
        observer_interpolater = interpolate.interp1d(all_angles, all_observer, axis=0, bounds_error=False,
                                                     fill_value='extrapolate')

        fov_bounds = [0, 0]

        model_angles = np.arange(all_angles[0] + fov_bounds[0],
                                 all_angles[-1] + fov_bounds[1] + model_vertical_resolution_rad,
                                 model_vertical_resolution_rad)

        for model_angle in model_angles:
            rot_matrix = util.rotation_matrix(rotation_axis, model_angle)

            look_vector = rot_matrix @ lower_los.look_vector

            mjd = np.asscalar(mjd_interpolator(model_angle))
            obs = observer_interpolater(model_angle)

            self.model_geometry.lines_of_sight.append(sk.LineOfSight(mjd, obs, look_vector))

        self.spectrographs = []

        all_vert_fov = []
        for idx, o in enumerate(self.measurement_optical_axis):
            vert_fov = float(self.measurement_l1.data.isel(los=idx).exposureTime.values) * self.scan_rate
            os_pixel = DeltaFunction()

            sensor = SpectrographFast(meas_wavel, os_pixel,
                                      vert_fov=Rectangle(width=vert_fov, mode='linear'),
                                      horiz_fov=Rectangle(width=1, mode='constant'))
            # sensor = SpectrographFast(model_wavel, os_pixel,
            #                           vert_fov=Gaussian(fwhm=0.0015502680433856196, mode='linear'),
            #                           horiz_fov=Rectangle(width=1, mode='constant'))

            self.spectrographs.append(sensor)

            all_vert_fov.append(vert_fov)

        self.model_wavel_nm = np.array(model_wavel)

        self.engine = sk.EngineHR(geometry=self.model_geometry, atmosphere=atmosphere, options=options)

        self.engine.wavelengths = self.model_wavel_nm

        self.engine.grid_spacing = 1000

        self._fwhm = osiris.skretrieval_l1().data['fwhm']

    def calculate_radiance(self):
        radiance = self.engine.calculate_radiance()
        if hasattr(radiance, 'weighting_function'):
            wf = radiance.weighting_function
            radiance = radiance.radiance
        else:
            wf = None

        model_values = [s.model_radiance(o, self.model_wavel_nm, self.model_geometry, radiance, wf) for o, s
                        in zip(self.measurement_optical_axis, self.spectrographs)]

        data = radianceformat.RadianceGridded(xr.concat([m.data for m in model_values], dim='los'))

        data.data['fwhm'] = self._fwhm

        return data


class IdealOSIRISModel(object):
    def __init__(self, osiris: Osiris, atmosphere: sk.Atmosphere, options=None, meas_wavel=None, mask_osiris_bad_pixels=True):
        self.measurement_l1 = osiris.skretrieval_l1()

        self.model_geometry = osiris.sasktran_geometry()

        self.model_wavel_nm = np.array(meas_wavel)

        self.engine = sk.EngineHR(geometry=self.model_geometry, atmosphere=atmosphere, options=options)

        self.engine.wavelengths = self.model_wavel_nm

        self.engine.grid_spacing = 1000

        self._fwhm = osiris.skretrieval_l1().data['fwhm']

        self._mask_osiris_bad_pixels = mask_osiris_bad_pixels

    def calculate_radiance(self):
        radiance = self.engine.calculate_radiance(output_format='xarray')

        if self.engine.atmosphere.wf_species is not None:
            data = radianceformat.RadianceGridded(radiance.rename({'wf_{}'.format(self.engine.atmosphere.wf_species): 'wf',
                                                                   'los_vector': 'los_vectors'}))
        else:
            data = radianceformat.RadianceGridded(radiance.rename({'los_vector': 'los_vectors'}))

        data.data['fwhm'] = self._fwhm

        if self._mask_osiris_bad_pixels:
            data.data['radiance'] = data.data['radiance'].where(~np.isnan(self.measurement_l1.data['radiance']))

        return data

if __name__ == "__main__":
    os = Osiris(41880012)

    os_model = OSIRISModel(os, sk.Atmosphere(), model_wavel=[350])