from osiris_skretrieval.osiris import Osiris
from osiris_skretrieval.osirismodel import OSIRISModel, IdealOSIRISModel
import sasktran as sk
from osiris_skretrieval.osiris_no2 import OSIRISNO2Retrieval
from osiris_skretrieval.osiris_o3 import OsirisOzoneRetrieval
from osiris_skretrieval.osiris_aero import OsirisAerosolRetrieval
from osiris_skretrieval.osiris_albedo import OSIRISAlbedoRetrieval
from osiris_skretrieval import data_folder
import numpy as np
from skretrieval.retrieval.rodgers import Rodgers
import xarray as xr
from osiris_skretrieval.osiris_spectral_fit import spectral_fit
from copy import copy
from osirisl1services.readlevel1 import open_level1_ecmwf
import julian
import logging
import yaml
from pathlib import Path
from osiris_skretrieval.cloud_detection import cloud_masks
from scipy import interpolate


class OsirisRetrieval(object):
    def __init__(self, scan_no, ret_order,
                 apply_rsas=True,
                 load_invalid_osiris_data=False,
                 osiris_l1_fit_psf=True,
                 anc_output_alts=np.linspace(500, 99500, 100),
                 include_o2o2=False,
                 no2_kwargs={},
                 o3_kwargs={},
                 aero_kwargs={},
                 albedo_kwargs={}):
        self._scan_no = scan_no
        self._ret_order = ret_order
        self._apply_rsas = apply_rsas
        self._load_invalid_osiris_data = load_invalid_osiris_data
        self._osiris_l1_fit_psf = osiris_l1_fit_psf
        self._anc_output_alts = anc_output_alts
        self._include_o2o2 = include_o2o2

        self._ret_function_map = {'no2': lambda x, y: self.retrieve_no2(x, y),
                                  'o3': lambda x, y: self.retrieve_o3(x, y),
                                  'aerosol': lambda x, y: self.retrieve_aero(x, y),
                                  'albedo': lambda x, y: self.retrieve_albedo(x, y)}

        self._set_no2_options(**no2_kwargs)
        self._set_o3_options(**o3_kwargs)
        self._set_aero_options(**aero_kwargs)
        self._set_albedo_options(**albedo_kwargs)

        self._aerosol_lowerbound = self._aero_options['lower_altitude_bound_m']

        self._cloud_altitude = 0
        self._psc_altitude = 0

        self._aerosol_reference_altitude = None

    def _set_no2_options(self, fwhm_type='spectral_fit', tikh_factor=100,
                         initial_no2='pratmo',
                         include_polarization_fit=False, retrieve_on_osiris_grid=False,
                         max_iter=10,
                         use_ideal_osiris=True,
                         lower_altitude_bound_m=8500,
                         upper_altitude_bound_m=48500,
                         high_alt_normalization_range=5000,
                         initial_lm_damping=0.5,
                         good_pixel_fraction=0.9999,
                         wavel_range=(434.7, 476.5),
                         use_cloud_lowerbound=False,
                         num_diffuse=1):
        self._no2_options = {'fwhm_type': fwhm_type,
                             'tikh_factor': tikh_factor,
                             'include_polarization_fit': include_polarization_fit,
                             'retrieve_on_osiris_grid': retrieve_on_osiris_grid,
                             'max_iter': max_iter,
                             'use_ideal_osiris': use_ideal_osiris,
                             'upper_altitude_bound_m': upper_altitude_bound_m,
                             'lower_altitude_bound_m': lower_altitude_bound_m,
                             'high_alt_normalization_range': high_alt_normalization_range,
                             'initial_lm_damping': initial_lm_damping,
                             'good_pixel_fraction': good_pixel_fraction,
                             'wavel_range': wavel_range,
                             'no2_rtm_alts': np.linspace(500, 99500, 100),
                             'initial_no2': initial_no2,
                             'use_cloud_lowerbound': use_cloud_lowerbound,
                             'num_diffuse': num_diffuse}

    def _set_o3_options(self, initial_ozone='labow', tikh_factor=20000, triplet_file='ozone_triplet_omps.yml',
                        upper_altitude_bound_m=59500,
                        lower_altitude_bound_m=8500,
                        state_vector_space='log',
                        ret_spacing_m=1000,
                        apriori_clim=None,
                        apriori_influence=0,
                        use_ideal_osiris=True,
                        use_cloud_lowerbound=False,
                        max_iter=15,
                        initial_lm_damping=0.5,
                        o3_cross='dbm',
                        num_diffuse=1,
                        use_full_normalization_range=True,
                        linear_combine_triplets=False,
                        rtm_507_replication=False,
                        retrieve_on_osiris_grid=False,
                        use_refraction=False):
        self._o3_options = {'initial_ozone': initial_ozone,
                            'o3_rtm_alts': np.linspace(500, 99500, 100),
                            'tikh_factor': tikh_factor,
                            'upper_altitude_bound_m': upper_altitude_bound_m,
                            'lower_altitude_bound_m': lower_altitude_bound_m,
                            'state_vector_space': state_vector_space,
                            'ret_spacing_m': ret_spacing_m,
                            'apriori_clim': apriori_clim,
                            'apriori_influence': apriori_influence,
                            'use_ideal_osiris': use_ideal_osiris,
                            'use_cloud_lowerbound': use_cloud_lowerbound,
                            'max_iter': max_iter,
                            'initial_lm_damping': initial_lm_damping,
                            'o3_cross': o3_cross,
                            'num_diffuse': num_diffuse,
                            'use_full_normalization_range': use_full_normalization_range,
                            'linear_combine_triplets': linear_combine_triplets,
                            'rtm_507_replication': rtm_507_replication,
                            'retrieve_on_osiris_grid': retrieve_on_osiris_grid,
                            'use_refraction': use_refraction}

        triplet_path = Path(triplet_file)
        if not triplet_path.is_absolute():
            triplet_path = data_folder().joinpath(triplet_path)

        self._o3_options['triplets'] = yaml.load(open(triplet_path.as_posix(), 'r'), Loader=yaml.FullLoader)['triplets']

    def _set_aero_options(self, initial_aero=None,
                          v7_folder=None,
                          tikh_factor=20000,
                          upper_altitude_bound_m=59500,
                          lower_altitude_bound_m=8500,
                          state_vector_space='log',
                          ret_spacing_m=1000,
                          apriori_clim=None,
                          apriori_influence=0,
                          use_ideal_osiris=True,
                          max_iter=15,
                          initial_lm_damping=0.5,
                          high_altitude_normalize_jacobian=False,
                          num_diffuse=1,
                          use_apriori_upperbound=False,
                          normalization_covariance=False,
                          require_all_wavelengths_valid=True,
                          retrieve_on_osiris_grid=False,
                          use_jacobian_lowerbound=True,
                          assume_error_fraction=None):
        self._aero_options = {'initial_aero': initial_aero,
                              'v7_folder': v7_folder,
                              'tikh_factor': tikh_factor,
                              'aero_rtm_alts': np.linspace(500, 99500, 100),
                              'upper_altitude_bound_m': upper_altitude_bound_m,
                              'lower_altitude_bound_m': lower_altitude_bound_m,
                              'state_vector_space': state_vector_space,
                              'ret_spacing_m': ret_spacing_m,
                              'apriori_clim': apriori_clim,
                              'apriori_influence': apriori_influence,
                              'use_ideal_osiris': use_ideal_osiris,
                              'max_iter': max_iter,
                              'initial_lm_damping': initial_lm_damping,
                              'high_altitude_normalize_jacobian': high_altitude_normalize_jacobian,
                              'num_diffuse': num_diffuse,
                              'use_apriori_upperbound': use_apriori_upperbound,
                              'normalization_covariance': normalization_covariance,
                              'require_all_wavelengths_valid': require_all_wavelengths_valid,
                              'retrieve_on_osiris_grid': retrieve_on_osiris_grid,
                              'use_jacobian_lowerbound': use_jacobian_lowerbound,
                              'assume_error_fraction': assume_error_fraction}

    def _set_albedo_options(self, initial_albedo=0.3,
                            max_iter=5,
                            initial_lm_damping=0.5,
                            retrieval_wavelength=675,
                            num_diffuse=1,
                            use_aerosol_reference_altitude=False):
        self._albedo_options = {'initial_albedo': initial_albedo,
                                'max_iter': max_iter,
                                'initial_lm_damping': initial_lm_damping,
                                'retrieval_wavelength': retrieval_wavelength,
                                'num_diffuse': num_diffuse,
                                'use_aerosol_reference_altitude': use_aerosol_reference_altitude}

    def load_osiris_data(self):
        logging.info(f'Loading OSIRIS Scan {self._scan_no}')
        os_data = Osiris(self._scan_no, load_psf=self._osiris_l1_fit_psf, apply_rsas=self._apply_rsas,
                         load_invalid=self._load_invalid_osiris_data)

        # If we are retrieving NO2 we do a second spectral fit for the NO2 absorption band
        if 'no2' in self._ret_order:
            if self._no2_options['fwhm_type'] == 'spectral_fit':
                spectral_fits = spectral_fit(os_data.skretrieval_l1(), fit_alt_range=(40000, 50000))

                no2_wavel_index = (os_data.spectrograph_data['wavelength'].values > self._no2_options['wavel_range'][0] - 5) & \
                                  (os_data.spectrograph_data['wavelength'].values < self._no2_options['wavel_range'][1] + 5)

                os_data.spectrograph_data['fwhm'].values[no2_wavel_index] = np.ones_like(os_data.spectrograph_data['fwhm'].values[no2_wavel_index]) * \
                                                           spectral_fits['fwhm']

        skl1 = os_data.skretrieval_l1()
        tl = os_data.skretrieval_l1().tangent_locations()

        self._mean_mjd = float(skl1.data.where(tl.altitude < 40000).mjd.mean())
        self._mean_lat = float(tl.where(tl.altitude < 40000).latitude.mean())
        self._mean_lon = float(tl.where(tl.altitude < 40000).longitude.mean())

        return os_data

    def initial_no2(self, os_data):
        no2_alts = self._no2_options['no2_rtm_alts']
        convolved_no2 = sk.OpticalPropertyConvolved(sk.NO2Vandaele1998(),
                                                    os_data.skretrieval_l1().data['wavelength'].values,
                                                    os_data.skretrieval_l1().data.fwhm.values / 2.355)
        if self._no2_options['initial_no2'] == 'pratmo':
            pratmo_sim = sk.Pratmo()

            no2_vals = pratmo_sim.get_parameter('SKCLIMATOLOGY_NO2_CM3', self._mean_lat, self._mean_lon, no2_alts,
                                                self._mean_mjd)


        elif self._no2_options['initial_no2'] is None:
            no2_vals = np.zeros_like(no2_alts)
        else:
            raise ValueError('initial_no2 must be one of pratmo or None')

        no2_user_def = sk.ClimatologyUserDefined(no2_alts, {'SKCLIMATOLOGY_NO2_CM3': no2_vals})

        return sk.Species(convolved_no2, no2_user_def)

    def initial_ozone(self, os_data):
        if self._o3_options['o3_cross'] == 'dbm':
            convolved_o3 = sk.OpticalPropertyConvolved(sk.O3DBM(), os_data.skretrieval_l1().data['wavelength'].values,
                                                       os_data.skretrieval_l1().data.fwhm.values / 2.355)
        elif self._o3_options['o3_cross'] == 'osiris':
            convolved_o3 = sk.O3OSIRISRes()
        else:
            raise ValueError('o3_cross must be one of dbm or osiris')

        o3_alts = self._o3_options['o3_rtm_alts']

        if self._o3_options['initial_ozone'] == 'labow':
            labow = sk.Labow()

            o3_vals = labow.get_parameter('SKCLIMATOLOGY_O3_CM3', self._mean_lat, self._mean_lon, o3_alts,
                                                self._mean_mjd)
        elif self._o3_options['initial_ozone'] is None:
            o3_vals = np.ones_like(o3_alts) * 0.0
        else:
            raise ValueError('o3_rtm_alts must be one of labow or None')

        o3_user_def = sk.ClimatologyUserDefined(o3_alts, {'SKCLIMATOLOGY_O3_CM3': o3_vals})
        return sk.Species(convolved_o3, o3_user_def)

    def background_atmosphere(self):
        ecmwf_data = open_level1_ecmwf(self._scan_no)
        pressure = ecmwf_data.density / 100 * ecmwf_data.temperature * 8.31446261815324 / 6.02214086e23 * 1e6

        background = sk.ClimatologyUserDefined(ecmwf_data.Altitude.values,
                                               {'SKCLIMATOLOGY_PRESSURE_PA': pressure.values * 100,
                                                'SKCLIMATOLOGY_AIRNUMBERDENSITY_CM3': ecmwf_data.density.values,
                                                'SKCLIMATOLOGY_TEMPERATURE_K': ecmwf_data.temperature.values})

        return sk.Species(sk.Rayleigh(), background, species='SKCLIMATOLOGY_AIRNUMBERDENSITY_CM3')

    def initial_aerosol(self):
        particle_size_dist = sk.ClimatologyUserDefined(altitudes=[0, 100000],
                                                       values={'SKCLIMATOLOGY_LOGNORMAL_MODEWIDTH': [1.6, 1.6],
                                                               'SKCLIMATOLOGY_LOGNORMAL_MODERADIUS_MICRONS': [0.08,
                                                                                                              0.08]})

        aero_opt_prop = sk.MieAerosol(particle_size_dist, 'H2SO4')
        xs_aero = aero_opt_prop.calculate_cross_sections(sk.MSIS90(), 0, 0, 20000, 54372, 750).total

        rtm_alts = self._aero_options['aero_rtm_alts']

        if self._aero_options['initial_aero'] == 'v7':
            if self._aero_options['v7_folder'] is not None:
                date = julian.from_jd(self._mean_mjd, fmt='mjd')
                aerosol_data = xr.open_mfdataset(
                    self._aero_options['v7_folder'] + '/AEROSOL-L2-LP-OSIRIS_ODIN-SASK_v7_0-{}{}.nc'.format(str(date.year).zfill(4),
                                                                                                  str(date.month).zfill(2)),
                    combine='by_coords', concat_dim='profile_id')

                aerosol_data = aerosol_data.sel(profile_id=self._scan_no)

                numden_aero = aerosol_data._extinction_rtm / xs_aero / 1e5

                rtm_aero = np.interp(rtm_alts, aerosol_data.altitude.values * 1000, numden_aero.values, left=numden_aero[0],
                                     right=0)
        elif self._aero_options['initial_aero'] is None:
            rtm_aero = np.ones_like(rtm_alts) * 0.0
        elif self._aero_options['initial_aero'] == 'v7_apriori':
            from osiris_skretrieval import data_folder
            data = np.loadtxt(data_folder().joinpath('aerosol_apriori.txt').as_posix())

            rtm_aero = np.interp(rtm_alts, data[:, 0], data[:, 1], left=data[0, 1], right=0)
            rtm_aero[rtm_aero==0] = 1e-20

        else:
            raise ValueError('initial_aero must be one of v7 or None')

        aero_clim = sk.ClimatologyUserDefined(rtm_alts, {'SKCLIMATOLOGY_AEROSOL_CM3': rtm_aero})

        return sk.Species(aero_opt_prop, aero_clim)

    def initial_atmosphere(self, os_data):
        atmo = sk.Atmosphere()

        o3_species = self.initial_ozone(os_data)
        if o3_species is not None:
            atmo['o3'] = o3_species

        air_species = self.background_atmosphere()
        atmo['air'] = air_species
        atmo.atmospheric_state = air_species.climatology

        aero_species = self.initial_aerosol()
        if aero_species is not None:
            atmo['aerosol'] = aero_species

        no2_species = self.initial_no2(os_data)
        if no2_species is not None:
            atmo['no2'] = no2_species

        atmo.brdf = self._albedo_options['initial_albedo']

        if self._include_o2o2:
            if 'SKCLIMATOLOGY_O2_CM3' in atmo.atmospheric_state.supported_species():
                o2_clim = atmo.atmospheric_state
            else:
                o2_clim = sk.MSIS90(includes=['o2'])

            o2 = o2_clim.get_parameter('SKCLIMATOLOGY_O2_CM3', self._mean_lat, self._mean_lon, self._o3_options['o3_rtm_alts'], self._mean_mjd)
            o2_clim_airdens = o2_clim.get_parameter('SKCLIMATOLOGY_AIRNUMBERDENSITY_CM3', self._mean_lat, self._mean_lon, self._o3_options['o3_rtm_alts'], self._mean_mjd)
            back_clim_airdens = atmo.atmospheric_state.get_parameter('SKCLIMATOLOGY_AIRNUMBERDENSITY_CM3', self._mean_lat, self._mean_lon, self._o3_options['o3_rtm_alts'], self._mean_mjd)
            # Scale the o2 in vmr
            o2 *= back_clim_airdens / o2_clim_airdens

            o2o2_clim = sk.ClimatologyUserDefined(self._o3_options['o3_rtm_alts'], {'o2o2': o2**2})

            o2o2_species = sk.Species(sk.OpticalProperty('O2_O2_HITRAN2016'), o2o2_clim)
            atmo['o2o2'] = o2o2_species

        return atmo

    def retrieve_no2(self, atmo_ret, os_data):
        logging.info('Starting NO2 Retrieval\n')
        if self._no2_options['use_cloud_lowerbound']:
            lower_bound = np.max([self._no2_options['lower_altitude_bound_m'], self._cloud_altitude, self._aerosol_lowerbound])
        else:
            lower_bound = self._no2_options['lower_altitude_bound_m']

        no2_ret = OSIRISNO2Retrieval(atmo_ret['no2'], os_data.skretrieval_l1(),
                                     tikh_factor=self._no2_options['tikh_factor'], high_alt_normalize=True,
                                     no2_opticalproperty=atmo_ret['no2'].optical_property,
                                     o3_opticalproperty=atmo_ret['o3'].optical_property,
                                     include_polarization_fit=self._no2_options['include_polarization_fit'],
                                     retrieve_on_osiris_grid=self._no2_options['retrieve_on_osiris_grid'],
                                     v6_normalization_altitude=False,
                                     high_alt_normalization_range=self._no2_options['high_alt_normalization_range'],
                                     upper_altitude_bound=self._no2_options['upper_altitude_bound_m'],
                                     good_pixel_fraction=self._no2_options['good_pixel_fraction'],
                                     lower_altitude_bound=lower_bound)

        no2_alts = self._no2_options['no2_rtm_alts']
        options, atmo_ret.wf_species = no2_ret.rtm_options()
        options['numdiffuseprofilesinplane'] = self._no2_options['num_diffuse']

        all_os_wavel = os_data.skretrieval_l1().data.wavelength.values
        os_wavel = all_os_wavel[(all_os_wavel > self._no2_options['wavel_range'][0]) & (all_os_wavel < self._no2_options['wavel_range'][1])]

        if self._no2_options['use_ideal_osiris']:
            os_model_ret = IdealOSIRISModel(os_data, atmo_ret, options, os_wavel)
        else:
            os_model_ret = OSIRISModel(os_data, atmo_ret, model_wavel=os_wavel, meas_wavel=os_wavel, options=options,
                                       model_vertical_resolution_rad=0.0002)

        ret = Rodgers(max_iter=self._no2_options['max_iter'],
                      lm_damping=self._no2_options['initial_lm_damping'], iterative_update_lm=True,
                      convergence_factor=1.01)
        ret_results = ret.retrieve(os_data.skretrieval_l1(), os_model_ret, no2_ret)

        ret_no2 = atmo_ret['no2'].climatology.get_parameter('SKCLIMATOLOGY_NO2_CM3', 0, 0, no2_alts, 54372)
        valid_no2 = copy(ret_no2)

        y_meas = no2_ret.measurement_vector(os_data.skretrieval_l1())
        low_alt = np.max([y_meas['y_altitudes'][~np.isnan(y_meas['y'])][0], lower_bound])
        high_alt = self._no2_options['upper_altitude_bound_m']

        valid_no2[no2_alts < low_alt] = np.nan
        valid_no2[no2_alts > high_alt] = np.nan

        error_covar = no2_ret.map_error_covariance(ret_results['error_covariance_from_noise'])

        ret_data = xr.Dataset({'no2': (['altitude'], ret_no2),
                               'valid_no2': (['altitude'], valid_no2),
                               'chi_sq': ret_results['chi_sq'],
                               'chi_sq_linear': ret_results['chi_sq_linear'],
                               'chi_sq_meas': ret_results['chi_sq_meas'],
                               'chi_sq_meas_linear': ret_results['chi_sq_meas_linear'],
                               'scanno': self._scan_no,
                               'mean_mjd': self._mean_mjd,
                               'mean_lat': self._mean_lat,
                               'mean_lon': self._mean_lon,
                               'error_covariance': (['altitude', 'altitude_2'], error_covar)},
                              coords={'altitude': no2_alts,
                                      'altitude_2': no2_alts})

        diag_data = xr.Dataset(
            {'averaging_kernel': (['ret_altitudes', 'true_altitudes'], ret_results['averaging_kernel']),
             'y_ret': (['y_altitude'], ret_results['ys'][-1]),
             'y_meas': (['y_altitude'], ret_results['y_meas']),
             'fit_residuals': (['y_altitude', 'fit_wavelength'], y_meas['fit_residuals']),
             'good_pixel_fraction': (['y_altitude'], y_meas['good_pixel_fraction'])},
            coords={'y_altitude': y_meas['y_altitudes'],
                    'fit_wavelength': y_meas['fit_residual_wavelengths'],
                    'ret_altitudes': no2_ret._retrieval_altitudes,
                    'true_altitudes': no2_ret._retrieval_altitudes})

        return ret_data, diag_data

    def retrieve_o3(self, atmo_ret, os_data):
        logging.info('Starting O3 Retrieval\n')
        if self._o3_options['use_cloud_lowerbound']:
            lower_bound = np.max([self._o3_options['lower_altitude_bound_m'], self._cloud_altitude, self._aerosol_lowerbound])
        else:
            lower_bound = self._o3_options['lower_altitude_bound_m']

        o3_ret = OsirisOzoneRetrieval(species=atmo_ret['o3'],
                                      measurement_l1=os_data.skretrieval_l1(),
                                      tikh_factor=self._o3_options['tikh_factor'],
                                      upper_altitude_bound_m=self._o3_options['upper_altitude_bound_m'],
                                      lower_altitude_bound_m=lower_bound,
                                      triplets=self._o3_options['triplets'],
                                      state_vector_space=self._o3_options['state_vector_space'],
                                      ret_spacing_m=self._o3_options['ret_spacing_m'],
                                      apriori_clim=self._o3_options['apriori_clim'],
                                      apriori_influence=self._o3_options['apriori_influence'],
                                      use_full_normalization_range=self._o3_options['use_full_normalization_range'],
                                      linear_combine_triplets=self._o3_options['linear_combine_triplets'],
                                      retrieve_on_osiris_grid=self._o3_options['retrieve_on_osiris_grid'])

        o3_alts = self._o3_options['o3_rtm_alts']
        options, atmo_ret.wf_species = o3_ret.rtm_options()
        options['numdiffuseprofilesinplane'] = self._o3_options['num_diffuse']

        if self._o3_options['use_refraction']:
            options['userefraction'] = 1

        if self._o3_options['rtm_507_replication']:
            options['integrationtechnique'] = 0
            options['forcev21diffuseincoming'] = 1

            if not self._o3_options['use_refraction']:
                options['useshellraytracer'] = 1

        os_wavel = o3_ret.measurement_wavelengths(os_data.skretrieval_l1())

        if self._o3_options['use_ideal_osiris']:
            os_model_ret = IdealOSIRISModel(os_data, atmo_ret, options, os_wavel)
        else:
            os_model_ret = OSIRISModel(os_data, atmo_ret, model_wavel=os_wavel, meas_wavel=os_wavel, options=options,
                                       model_vertical_resolution_rad=0.0002)

        ret = Rodgers(max_iter=self._o3_options['max_iter'],
                      lm_damping=self._o3_options['initial_lm_damping'], iterative_update_lm=True,
                      convergence_factor=1.01)
        ret_results = ret.retrieve(os_data.skretrieval_l1(), os_model_ret, o3_ret)

        ret_o3 = atmo_ret['o3'].climatology.get_parameter('SKCLIMATOLOGY_O3_CM3', 0, 0, o3_alts, 54372)
        valid_o3 = copy(ret_o3)

        y_meas = o3_ret.measurement_vector(os_data.skretrieval_l1())
        low_alt = o3_ret._retrieval_altitudes[0]
        high_alt = o3_ret._retrieval_altitudes[-1]

        valid_o3[o3_alts < low_alt] = np.nan
        valid_o3[o3_alts > high_alt] = np.nan

        error_covar = o3_ret.map_error_covariance(ret_results['error_covariance_from_noise'])

        ret_data = xr.Dataset({'o3': (['altitude'], ret_o3),
                               'valid_o3': (['altitude'], valid_o3),
                               'chi_sq': ret_results['chi_sq'],
                               'chi_sq_linear': ret_results['chi_sq_linear'],
                               'chi_sq_meas': ret_results['chi_sq_meas'],
                               'chi_sq_meas_linear': ret_results['chi_sq_meas_linear'],
                               'scanno': self._scan_no,
                               'mean_mjd': self._mean_mjd,
                               'mean_lat': self._mean_lat,
                               'mean_lon': self._mean_lon,
                               'error_covariance': (['altitude', 'altitude_2'], error_covar)},
                              coords={'altitude': o3_alts,
                                      'altitude_2': o3_alts})

        diag_data = xr.Dataset(
            {'averaging_kernel': (['ret_altitudes', 'true_altitudes'], ret_results['averaging_kernel'])},
            coords={'ret_altitudes': o3_ret._retrieval_altitudes,
                    'true_altitudes': o3_ret._retrieval_altitudes})

        return ret_data, diag_data

    def retrieve_aero(self, atmo_ret, os_data):
        logging.info('Starting Aerosol Retrieval\n')
        aero_ret = OsirisAerosolRetrieval(species=atmo_ret['aerosol'],
                                          psize_clim=atmo_ret['aerosol'].optical_property.particlesize_climatology,
                                          measurement_l1=os_data.skretrieval_l1(),
                                          tikh_factor=self._aero_options['tikh_factor'],
                                          upper_altitude_bound_m=self._aero_options['upper_altitude_bound_m'],
                                          lower_altitude_bound_m=self._aerosol_lowerbound,
                                          state_vector_space=self._aero_options['state_vector_space'],
                                          ret_spacing_m=self._aero_options['ret_spacing_m'],
                                          apriori_clim=self._aero_options['apriori_clim'],
                                          apriori_influence=self._aero_options['apriori_influence'],
                                          high_altitude_normalize_jacobian=self._aero_options['high_altitude_normalize_jacobian'],
                                          use_apriori_for_upperbound=self._aero_options['use_apriori_upperbound'],
                                          normalization_covariance=self._aero_options['normalization_covariance'],
                                          require_all_wavelengths_valid=self._aero_options['require_all_wavelengths_valid'],
                                          retrieve_on_osiris_grid=self._aero_options['retrieve_on_osiris_grid'],
                                          use_jacobian_for_lowerbound=self._aero_options['use_jacobian_lowerbound'],
                                          assume_error_fraction=self._aero_options['assume_error_fraction'])

        aero_alts = self._aero_options['aero_rtm_alts']
        options, atmo_ret.wf_species = aero_ret.rtm_options()
        options['numdiffuseprofilesinplane'] = self._aero_options['num_diffuse']

        os_wavel = aero_ret.measurement_wavelengths(os_data.skretrieval_l1())

        if self._aero_options['use_ideal_osiris']:
            os_model_ret = IdealOSIRISModel(os_data, atmo_ret, options, os_wavel)
        else:
            os_model_ret = OSIRISModel(os_data, atmo_ret, model_wavel=os_wavel, meas_wavel=os_wavel, options=options,
                                       model_vertical_resolution_rad=0.0002)

        aero_ret.optimize_settings(os_model_ret, os_data.skretrieval_l1(True))

        ret = Rodgers(max_iter=self._aero_options['max_iter'],
                      lm_damping=self._aero_options['initial_lm_damping'], iterative_update_lm=True,
                      convergence_factor=1.01)
        ret_results = ret.retrieve(os_data.skretrieval_l1(True), os_model_ret, aero_ret)

        ret_aero = atmo_ret['aerosol'].climatology.get_parameter('SKCLIMATOLOGY_AEROSOL_CM3', 0, 0, aero_alts, 54372)
        valid_aero = copy(ret_aero)

        y_meas = aero_ret.measurement_vector(os_data.skretrieval_l1())
        low_alt = aero_ret._retrieval_altitudes[0]
        high_alt = aero_ret._retrieval_altitudes[-1]

        valid_aero[aero_alts < low_alt] = np.nan
        valid_aero[aero_alts > high_alt] = np.nan

        error_covar = aero_ret.map_error_covariance(ret_results['error_covariance_from_noise'])

        xs = atmo_ret['aerosol'].optical_property.calculate_cross_sections(sk.MSIS90(), 0, 0, 20000, 54372, aero_ret._extinction_wavelength).total

        triplet = aero_ret._triplets[0]
        interp_weights = interpolate.interp1d(triplet['weight_altitudes'], triplet['weights'], axis=1, bounds_error=False, fill_value=np.nan)(aero_alts)

        ret_data = xr.Dataset({'aero_numden': (['altitude'], ret_aero),
                               'valid_aero_numden': (['altitude'], valid_aero),
                               'aero_extinction': (['altitude'], ret_aero * xs * 1e5),
                               'valid_aero_extinction': (['altitude'], valid_aero * xs * 1e5),
                               'wavelength_weights': (['wavelength', 'altitude'], interp_weights),
                               'chi_sq': ret_results['chi_sq'],
                               'chi_sq_linear': ret_results['chi_sq_linear'],
                               'chi_sq_meas': ret_results['chi_sq_meas'],
                               'chi_sq_meas_linear': ret_results['chi_sq_meas_linear'],
                               'scanno': self._scan_no,
                               'mean_mjd': self._mean_mjd,
                               'mean_lat': self._mean_lat,
                               'mean_lon': self._mean_lon,
                               'error_covariance': (['altitude', 'altitude_2'], error_covar * (xs * 1e5)**2)},
                              coords={'altitude': aero_alts,
                                      'altitude_2': aero_alts,
                                      'wavelength': triplet['wavelengths']})

        anc_ds = self.ancillary_ds(atmo_ret)
        # Retrieve cloud altitude
        cloud_altitude, psc_altitude, dGdlambda = cloud_masks(ret_data['aero_extinction'], anc_ds['temperature'],
                                                              os_data.skretrieval_l1(),
                                                              self._mean_lat)

        ret_data['cloud_altitude'] = cloud_altitude
        ret_data['psc_altitude'] = psc_altitude

        self._cloud_altitude = cloud_altitude
        self._psc_altitude = psc_altitude

        # Update the lowerbound since it could have changed due to the jacobian
        self._aerosol_lowerbound = aero_ret._retrieval_altitudes[0]

        # Update the reference altitude in case the albedo retrieval wants to use it later
        self._aerosol_reference_altitude = aero_ret._triplets[0]['normalization_range'][0]

        diag_data = xr.Dataset(
            {'averaging_kernel': (['ret_altitudes', 'true_altitudes'], ret_results['averaging_kernel']),
             'dGdlambda': (['altitude'], dGdlambda.values[:, 0]),
             'y_ret': (['y_altitude'], ret_results['ys'][-1]),
             'y_meas': (['y_altitude'], ret_results['y_meas'])
             },
            coords={'y_altitude': y_meas['y_altitude'],
                    'ret_altitudes': aero_ret._retrieval_altitudes,
                    'true_altitudes': aero_ret._retrieval_altitudes,
                    'altitude': dGdlambda.altitude.values})

        return ret_data, diag_data

    def retrieve_albedo(self, atmo_ret, os_data):
        logging.info('Starting Albedo Retrieval\n')

        albedo_ret = OSIRISAlbedoRetrieval(atmo_ret, 40000, self._albedo_options['retrieval_wavelength'],
                                           use_aerosol_reference_altitude=self._albedo_options['use_aerosol_reference_altitude'])

        os_wavel = albedo_ret.measurement_wavelengths(os_data.skretrieval_l1())

        options = albedo_ret.rtm_options()
        options['numdiffuseprofilesinplane'] = self._albedo_options['num_diffuse']

        if self._aero_options['use_ideal_osiris']:
            os_model_ret = IdealOSIRISModel(os_data, atmo_ret, options, os_wavel)
        else:
            os_model_ret = OSIRISModel(os_data, atmo_ret, model_wavel=os_wavel, meas_wavel=os_wavel, options=options,
                                       model_vertical_resolution_rad=0.0002)

        albedo_ret.configure_from_model(os_model_ret, os_data.skretrieval_l1(True))

        ret = Rodgers(max_iter=self._albedo_options['max_iter'],
                      lm_damping=self._albedo_options['initial_lm_damping'], iterative_update_lm=True,
                      convergence_factor=1.01)
        ret_results = ret.retrieve(os_data.skretrieval_l1(remove_sun=True), os_model_ret, albedo_ret)

        ret_albedo = atmo_ret.brdf.albedo

        ret_data = xr.Dataset({'albedo': ret_albedo,
                               'chi_sq': ret_results['chi_sq'],
                               'chi_sq_linear': ret_results['chi_sq_linear'],
                               'chi_sq_meas': ret_results['chi_sq_meas'],
                               'chi_sq_meas_linear': ret_results['chi_sq_meas_linear'],
                               'scanno': self._scan_no,
                               'mean_mjd': self._mean_mjd,
                               'mean_lat': self._mean_lat,
                               'mean_lon': self._mean_lon})

        return ret_data, None

    def ancillary_ds(self, atmo_ret):
        temperature = atmo_ret.atmospheric_state.get_parameter('SKCLIMATOLOGY_TEMPERATURE_K', self._mean_lat,
                                                               self._mean_lon,
                                                               self._anc_output_alts,
                                                               self._mean_mjd)

        pressure = atmo_ret.atmospheric_state.get_parameter('SKCLIMATOLOGY_PRESSURE_PA', self._mean_lat,
                                                            self._mean_lon,
                                                            self._anc_output_alts,
                                                            self._mean_mjd)

        air_density = atmo_ret['air'].climatology.get_parameter('SKCLIMATOLOGY_AIRNUMBERDENSITY_CM3', self._mean_lat,
                                                                self._mean_lon,
                                                                self._anc_output_alts,
                                                                self._mean_mjd)

        return xr.Dataset({'temperature': (['altitude'], temperature),
                                           'pressure': (['altitude'], pressure),
                                           'air_density': (['altitude'], air_density),
                           'scanno': self._scan_no},
                                         coords={'altitude': self._anc_output_alts})

    def retrieve(self):
        os_data = self.load_osiris_data()

        atmo_ret = self.initial_atmosphere(os_data)
        result = dict()

        result['ancillary'] = (self.ancillary_ds(atmo_ret), xr.Dataset())

        for target in self._ret_order:
            result[target] = self._ret_function_map[target](atmo_ret, os_data)

        return result

